-module(ccv_cvs_lib).

%-export([admin/1]).
%-export([test/1]).
-export([parse_file/3]).

-export([modified_date/2,
	 version/2]).


-include("ccv_locale.hrl").

%-define(dbg(Fmt, Args), io:fwrite("~p: " ++ Fmt, [?LINE|Args])).
-define(dbg(Fmt, Args), no_debug).

-record(state, {fd,
		userfun,
		userstate,
		chunksize = 100}).
		





parse_file(File, UserFun, UserState) ->
    {ok, Fd} = file:open(filename(File), [read, raw, binary]),
    S = #state{fd = Fd, userfun = UserFun, userstate = UserState},
    parse(chunk(S, <<>>), S).



modified_date(File, Label) ->
    {_Vsn, Date} = extract_info(File, Label),
    Date.

version(File, Label) ->
    {Vsn, _Date} = extract_info(File, Label),
    Vsn.

extract_info(File, Label) ->
    Fun = fun([], desc, _, {LabelVsn, [{_, Date}|_] = _Dates}) ->
		  {done, {LabelVsn, Date}};
	     (admin, symbols, Syms, _S) ->
		  {value, {_, LabelVsn}} = lists:keysearch(Label, 1, Syms),
		  {continue, {LabelVsn, []}};
	     (delta, num, Vsn, {LabelVsn, Dates}) ->
		  case lists:prefix(Vsn, LabelVsn) of
		      true ->
			  {continue, {LabelVsn, {match, Vsn, Dates}}};
		      false ->
			  {continue, {LabelVsn, Dates}}
		  end;
	     (delta, date, Date, {LabelVsn, {match, Vsn, Dates}}) ->
		  {continue, {LabelVsn, sort_insert(Vsn, Dates, Date)}};
	     (_Block, _Labeltag, _Value, S) ->
		  {continue, S}
	  end,
    parse_file(File, Fun, {undefined, undefined}).


sort_insert(K, [{K1,  _} = H|T], V) when K > K1 -> [{K, V}, H|T];
sort_insert(K, [H|T], V) ->			   [H|sort_insert(K, T, V)];
sort_insert(K, [],    V) ->			   [{K,V}].


filename(#ccv_file{type = ccv_cvs,
		   internal = {Root, _Tag, Fname}}) ->
    filename:join(Root, Fname ++ ",v");
filename(File) when list(File) ->
    File.



parse(Bin, S0) ->
    Grammar = 
	[{admin, once, 
	  [{head,    <<"head">>,    req,  fun(B,S) -> a_head(B,S) end},
	   {branch,  <<"branch">>,  '?',  fun(B,S) -> a_branch(B,S) end},
	   {access,  <<"access">>,  req,  fun(B,S) -> a_access(B,S) end},
	   {symbols, <<"symbols">>, req,  fun(B,S) -> a_symbols(B,S) end},
	   {locks,   <<"locks">>,   req,  fun(B,S) -> a_locks(B,S) end},
	   {strict,  <<"strict">>,  '?',  fun(B,S) -> a_strict(B,S) end},
	   {comment, <<"comment">>, '?',  fun(B,S) -> a_comment(B,S) end},
	   {expand,  <<"expand">>,  '?',  fun(B,S) -> a_expand(B,S) end},
	   newphrase]},
	 {delta, '*', is_num, 
	  [{num,     is_num,          req, fun(B,S) -> d_num(B,S) end},
	   {date,     <<"date">>,     req, fun(B,S) -> d_date(B,S) end},
	   {author,   <<"author">>,   req, fun(B,S) -> d_author(B,S) end},
	   {state,    <<"state">>,    req, fun(B,S) -> d_state(B,S) end},
	   {branches, <<"branches">>, req, fun(B,S) -> d_branches(B,S) end},
	   {next,     <<"next">>,     req, fun(B,S) -> d_next(B,S) end},
	   newphrase]},
	 {desc, <<"desc">>, once,  fun(B,S) -> desc(B,S) end},
	 {deltatext, '*', is_num, 
	  [{num, is_num,     req, fun(B,S) -> d_num(B,S) end},
	   {log, <<"log">>,  req, fun(B,S) -> d_log(B,S) end},
	   newphrase,
	   {text, <<"text">>, req, fun(B,S) -> deltatext(B,S) end}]}],
    fold(Grammar, Bin, S0).

%% --------------------------
%% individual parse functions

a_head(B, S) -> split_at(B, $;, S).
a_branch(B, S) -> split_at(B, $;, S).
a_access(B, S) -> split_at(B, $;, S).
a_symbols(B, S) -> parse_symbols(B, S).
a_locks(B, S) -> split_at(B, $;, S).
a_strict(B, S) -> split_at(B, $;, S).
a_comment(B, S) -> string_plus_semi(B, S).
a_expand(B, S) -> string_plus_semi(B, S).

d_date(B, S) -> parse_date(B, S).
d_author(B, S) -> split_at(B, $;, S).
d_state(B, S) -> split_at(B, $;, S).
d_branches(B, S) -> split_at(B, $;, S).
d_next(B, S) -> split_at(B, $;, S).


desc(B, S) -> parse_string(B, S).

d_num(B, S) -> opt_extract_num(B, S).
d_log(B, S) -> parse_string(B, S).
deltatext(B, S) -> parse_string(B, S).

%% end individual parse functions
%% ------------------------------


fold(Rules, B, S) when size(B) < 60 ->
    case chunk(S) of
	{ok, Bin} ->
	    fold1(Rules, list_to_binary([B, Bin]), S);
	eof ->
	    fold1(Rules, B, S)
    end;
fold(Rules, B, S) ->
    fold1(Rules, B, S).


fold1([{BlockLabel, once, Rules} = _R|T], B0, S0) ->
%    ?dbg("fold(~p, B0 = '~s...'~n", [R, bin2string(B0)]),
    case catch fold_block(Rules, B0, BlockLabel, T, S0) of
	{'EXIT', Reason} ->
	    exit(Reason);
	{done, Result} ->
	    Result;
	{B1, #state{} = S1} when binary(B1) ->
	    fold(T, B1, S1)
    end;
fold1([{BlockLabel, '*', MatchRule, Rules}|T] = Gram, B0, S0) ->
    %% 
    case match_rule(MatchRule, B0) of
	true ->
	    case catch fold_opt_block(Rules, B0, BlockLabel, 
				      MatchRule, T, S0) of
		{'EXIT', Reason} ->
		    exit(Reason);
		{done, Result} ->
		    Result;
		{B1, #state{} = S1} when binary(B1) ->
		    fold(Gram, B1, S1)
	    end;
	false ->
	    fold(T, B0, S0)
    end;
fold1([{Label, Pat, Occurrence, ParseFun}|T], B0, S0) when binary(Pat) ->
    Size = size(Pat),
    case {B0, Occurrence} of
	{<<Pat:Size/binary, Rest/binary>>, _} ->
	    ?dbg("Pat (~s) matched~n", [binary_to_list(Pat)]),
	    {B1, Value} = ParseFun(strip_ws(Rest, S0), S0),
	    case (S0#state.userfun)([], Label, Value, S0#state.userstate) of
		{done, UserState} ->
		    UserState;    % we're done
		{continue, UserState} ->
		    S1 = S0#state{userstate = UserState},
		    fold(T, strip_ws(B1, S1), S1)
	    end;
	{_, once} ->
	    exit({required, Label});
	_ ->
	    fold(T, B0, S0)
    end;
fold1([], _, S) ->
    S#state.userstate.


fold_block(Rules, B, BlockLabel, T0, S) when size(B) < 60 ->
    case chunk(S) of
	{ok, Bin} ->
	    fold_block1(Rules, list_to_binary([B, Bin]), BlockLabel, T0, S);
	eof ->
	    fold_block1(Rules, B, BlockLabel, T0, S)
    end;
fold_block(Rules, B, BlockLabel, T0, S) ->
    fold_block1(Rules, B, BlockLabel, T0, S).

fold_block1([{Label, Pat, Opt, Fun} = _R|T], B0, 
	   BlockLabel, T0, S0) when binary(Pat) ->
%    ?dbg("fold_block(~p, B0 = '~s...'~n", [R, bin2string(B0)]),
    Size = size(Pat),
    case {B0, Opt} of
	{<<Pat:Size/binary, Rest/binary>>, _} ->
	    ?dbg("Pat (~s) matched~n", [binary_to_list(Pat)]),
	    {B1, Value} = Fun(strip_ws(Rest, S0), S0),
	    F = S0#state.userfun,
	    ?dbg("calling ~p(~p,~p,~p,~p)~n", 
		      [F, BlockLabel, Label, Value, S0#state.userstate]),
	    case F(BlockLabel, Label, Value, S0#state.userstate) of
		{done, _UserState} = Res ->
		    throw(Res);
		{continue, UserState} ->
		    S1 = S0#state{userstate = UserState},
		    fold_block(T, strip_ws(B1, S1), BlockLabel, T0, S1)
	    end;
	{_, req} ->
	    exit({required, Label});
	_ ->
	    fold_block(T, B0, BlockLabel, T0, S0)
    end;
fold_block1([{Label, is_num, Opt}|T], B0, BlockLabel, T0, S0) ->
    case {opt_extract_num(B0, S0), Opt} of
	{false, req} ->
	    exit({required, Label});
	{{B1, Num}, _} ->
	    F = S0#state.userfun,
	    case F(BlockLabel, Label, Num, S0#state.userstate) of
		{done, _UserState} = Res ->
		    throw(Res);
		{continue, UserState} ->
		    S1 = S0#state{userstate = UserState},
		    fold_block(T, strip_ws(B1, S1), BlockLabel, T0, S1)
	    end
    end;
fold_block1([newphrase] = R, B0, BlockLabel, T0, S0) ->
    %% in this case, the next item either matches T0's next rule,
    %% or is a "newphrase"
    case peekahead(T0, B0, S0) of
	false ->
	    {B1, Label, Value} = newphrase(B0, S0),
	    F = S0#state.userfun,
	    case F(BlockLabel, Label, Value, S0#state.userstate) of
		{done, _UserState} = Res ->
		    throw(Res);
		{continue, UserState} ->
		    S1 = S0#state{userstate = UserState},
		    fold_block(R, strip_ws(B1, S1), BlockLabel, T0, S1)
	    end;
	true ->
	    {B0, S0}
    end;
fold_block1([], B0, _BlockLabel, _T0, S) ->
    {B0, S}.


fold_opt_block(Rules, B, BlockLabel, R0, T0, S) when size(B) < 60 ->
    case chunk(S) of
	{ok, Bin} ->
	    fold_opt_block1(Rules, list_to_binary([B, Bin]), 
			    BlockLabel, R0, T0, S);
	eof ->
	    fold_opt_block1(Rules, B, BlockLabel, R0, T0, S)
    end;
fold_opt_block(Rules, B, BlockLabel, R0, T0, S) ->
    fold_opt_block1(Rules, B, BlockLabel, R0, T0, S).


fold_opt_block1([{Label, Pat, Opt, Fun} = _R|T],
	       B0, BlockLabel, R0, T0, S0) when binary(Pat) ->
%    ?dbg("fold_opt_block(~p, B0 = '~s...'~n",
%	      [R, bin2string(B0)]),
    Size = size(Pat),
    case {B0, Opt} of
	{<<Pat:Size/binary, Rest/binary>>, _} ->
	    ?dbg("Pat (~s) matched~n", [binary_to_list(Pat)]),
	    {B1, Value} = Fun(strip_ws(Rest, S0), S0),
	    case (S0#state.userfun)(BlockLabel, Label, Value, 
				    S0#state.userstate) of
		{done, _UserState} = Res ->
		    throw(Res);
		{continue, UserState} ->
		    S1 = S0#state{userstate = UserState},
		    fold_opt_block(T, strip_ws(B1, S1), BlockLabel, R0, T0, S1)
	    end;
	{_, req} ->
	    exit({required, Label});
	_ ->
	    fold_opt_block(T, B0, BlockLabel, R0, T0, S0)
    end;
fold_opt_block1([{Label, is_num, Opt, Fun} = _R|T], B0, 
	       BlockLabel, R0, T0, S0) ->
%    ?dbg("fold_opt_block(~p, B0 = '~s...'~n",
%	      [R, bin2string(B0)]),
    case {Fun(B0, S0), Opt} of
	{false, req} ->
	    exit({required, Label});
	{{B1, Num}, _} ->
	    case (S0#state.userfun)(BlockLabel, Label, Num, 
				    S0#state.userstate) of
		{done, _UserState} = Res ->
		    throw(Res);
		{continue, UserState} ->
		    S1 = S0#state{userstate = UserState},
		    fold_opt_block(T, strip_ws(B1, S1), BlockLabel, R0, T0, S1)
	    end
    end;
fold_opt_block1([newphrase] = R, B0, BlockLabel, R0, T0, S0) ->
    %% in this case, the next item either matches T0's next rule,
    %% R0 (the MatchRule for a repeating block) or is a "newphrase"
    ?dbg("fold_opt_block(~p, B0 = '~s...'~n",
	      [R, bin2string(B0)]),
    case peekahead(T0, B0, S0) of
	false ->
	    case match_rule(R0, B0) of
		true ->
		    ?dbg("the same type of block again...~n", []),
		    {B0, S0};
		false ->
		    ?dbg("peekahead() -> false~n", []),
		    {B1, Label, Value} = newphrase(B0, S0),
		    case (S0#state.userfun)(BlockLabel, Label, Value, 
					    S0#state.userstate) of
			{done, _UserState} = Res ->
			    throw(Res);
			{continue, UserState} ->
			    S1 = S0#state{userstate = UserState},
			    fold_opt_block(R, strip_ws(B1, S1), 
					   BlockLabel, R0, T0, S1)
		    end
	    end;
	true ->
	    ?dbg("peekahead() -> true~n", []),
	    {B0, S0}
    end;
fold_opt_block1([newphrase|T] = Rs, B0, BlockLabel, R0, T0, S0) ->
    %% in this case, the next item either matches T's next rule,
    %% or is a "newphrase"
    ?dbg("fold_opt_block([newphrase|T], B0 = '~s...'~n",
	      [bin2string(B0)]),
    case peekahead(T, B0, S0) of
	false ->
	    {B1, Label, Value} = newphrase(B0, S0),
	    case (S0#state.userfun)(BlockLabel, Label, Value, 
				    S0#state.userstate) of
		{done, _UserState} = Res ->
		    throw(Res);
		{continue, UserState} ->
		    S1 = S0#state{userstate = UserState},
		    fold_opt_block(Rs, strip_ws(B1, S1), BlockLabel, 
				   R0, T0, S1)
	    end;
	true ->
	    fold_opt_block(T, B0, BlockLabel, R0, T0, S0)
    end;
fold_opt_block1([], B0, _BlockLabel, _R0, _T0, S) ->
    {B0, S}.


match_rule(is_num, <<H:8, _T/binary>>) when H>=$0, H=<$9 -> true;
match_rule(_, _) ->
    false.

peekahead([{_, '*', is_num, _}|_], B0, S0) ->
    ?dbg("peekahead: opt_extract_num('~s')~n", 
	      [bin2string(B0)]),
    case opt_extract_num(B0, S0) of
	false ->
	    false;
	_ ->
	    true
    end;
peekahead([{_Label, Pat, _, _}|_], B0, _S0) when binary(Pat) ->
    ?dbg("peekahead: match '~s' - '~s'~n",
	      [binary_to_list(Pat), bin2string(B0)]),
    Size = size(Pat),
    case B0 of
	<<Pat:Size/binary, _Rest/binary>> ->
	    true;
	_ ->
	    false
    end;
peekahead(_, _, _) ->
    false.

opt_extract_num(<<H:8, T/binary>>, S) when H >= $0, H =< $9 ->
    opt_extract_num(T, S, [H]);
opt_extract_num(_, _) ->
    false.

opt_extract_num(<<>>, S, Acc) ->
    case chunk(S) of
	{ok, Bin} ->
	    opt_extract_num1(Bin, S, Acc);
	eof ->
	    {<<>>, lists:reverse(Acc)}
    end;
opt_extract_num(B, S, Acc) ->
    opt_extract_num1(B, S, Acc).

opt_extract_num1(<<".", T/binary>>, S, Acc) ->
    opt_extract_num(T, S, "." ++ Acc);
opt_extract_num1(<<H:8, T/binary>>, S, Acc) when H >= $0, H =< $9 ->
    opt_extract_num(T, S, [H|Acc]);
opt_extract_num1(<<" ", T/binary>>, _S, Acc) ->
    {T, lists:reverse(Acc)};
opt_extract_num1(<<"\n", T/binary>>, _S, Acc) ->
    {T, lists:reverse(Acc)};
opt_extract_num1(<<"\t", T/binary>>, _S, Acc) ->
    {T, lists:reverse(Acc)};
opt_extract_num1(_Bin, _S, _Acc) ->
    false.


newphrase(B0, S) ->
    {B1, Label} = split_at_ws(B0, S),
    {B2, Value} = 
	case strip_ws(B1, S) of
	    <<"@", B11/binary>> ->
		{B12, _Str} = parse_string(B11, S, []),
		{B12, _} = split_at(B12, $;, S);
	    B21 ->
		split_at(B21, $;, S)
	end,
    {B2, Label, Value}.




chunk(S, Sofar) ->
    case chunk(S) of
	{ok, Bin} ->
	    list_to_binary([Sofar, Bin]);
	eof ->
	    Sofar
    end.

chunk(#state{fd = Fd, chunksize = Sz}) ->
    ?dbg(">>> chunk <<<<~n", []),
    case file:read(Fd, Sz) of
	{error, Reason} ->
	    exit(Reason);
	Other ->
	    Other
    end.

strip_ws(<<>>, S) ->
    case chunk(S) of
	{ok, Bin} ->
	    strip_ws1(Bin, S);
	eof ->
	    list_to_binary([])  % my emacs couldn't indent after <<>> 
    end;
strip_ws(B, S) ->
    strip_ws1(B, S).

strip_ws1(<<" ",Tail/binary>>, S) ->
    strip_ws(Tail, S);
strip_ws1(<<"\t",Tail/binary>>, S) ->
    strip_ws(Tail, S);
strip_ws1(<<"\n",Tail/binary>>, S) ->
    strip_ws(Tail, S);
strip_ws1(Bin, _S) ->
    Bin.



parse_symbols(Bin, S) ->
    parse_symbols(Bin, S, []).

parse_symbols(<<";", T/binary>>, _S, Acc) ->
    {T, lists:reverse(Acc)};
parse_symbols(B0, S, Acc) ->
    B1 = strip_ws(B0, S),
    {B2, Sym} = split_at(B1, $:, S),
    {B3, Id} = split_at_ws_or_semi(B2, S),
    parse_symbols(B3, S, [{Sym, Id}|Acc]).


split_at(Bin, Char, S) ->
    split_at(Bin, Char, S, []).

split_at(<<>>, H, S, Acc) ->
    {ok, Bin} = chunk(S),
    split_at1(Bin, H, S, Acc);
split_at(B, H, S, Acc) ->
    split_at1(B, H, S, Acc).

split_at1(<<H:8, T/binary>>, H, _S, Acc) ->
    {T, lists:reverse(Acc)};
split_at1(<<H:8, T/binary>>, X, S, Acc) ->
    split_at(T, X, S, [H|Acc]).



%%
split_at_ws_or_semi(Bin, S) ->
    split_at_ws_or_semi(Bin, S, []).


split_at_ws_or_semi(<<>>, S, Acc) ->
    {ok, Bin} = chunk(S),
    split_at_ws_or_semi1(Bin, S, Acc);
split_at_ws_or_semi(B, S, Acc) ->
    split_at_ws_or_semi1(B, S, Acc).

split_at_ws_or_semi1(<<";", _T/binary>> = Bin, _S, Acc) ->
    {Bin, lists:reverse(Acc)};
split_at_ws_or_semi1(<<" ", T/binary>>, _S, Acc) ->
    {T, lists:reverse(Acc)};
split_at_ws_or_semi1(<<"\t", T/binary>>, _S, Acc) ->
    {T, lists:reverse(Acc)};
split_at_ws_or_semi1(<<"\n", T/binary>>, _S, Acc) ->
    {T, lists:reverse(Acc)};
split_at_ws_or_semi1(<<H:8, T/binary>>, S, Acc) ->
    split_at_ws_or_semi(T, S, [H|Acc]).


%%
split_at_ws(Bin, S) ->
    split_at_ws(Bin, S, []).

split_at_ws(<<>>, S, Acc) ->
    case chunk(S) of
	{ok, Bin} ->
	    split_at_ws1(Bin, S, Acc);
	eof ->
	    {lists:reverse(Acc), <<>>}
    end;
split_at_ws(B, S, Acc) ->
    split_at_ws1(B, S, Acc).

split_at_ws1(<<" ", T/binary>>, _S, Acc) ->
    {lists:reverse(Acc), T};
split_at_ws1(<<"\t", T/binary>>, _S, Acc) ->
    {lists:reverse(Acc), T};
split_at_ws1(<<"\n", T/binary>>, _S, Acc) ->
    {lists:reverse(Acc), T};
split_at_ws1(<<H:8, T/binary>>, S, Acc) ->
    split_at_ws(T, S, [H|Acc]).




parse_date(B, S) when size(B) < 17 ->
    case chunk(S) of
	{ok, Bin} ->
	    parse_date1(list_to_binary([B, Bin]), S);
	eof ->
	    parse_date1(B, S)
    end;
parse_date(B, S) ->
    parse_date1(B, S).

parse_date1(<<Y1:8,Y2:8, ".", M1:8,M2:8,   ".", D1:8,D2:8, ".",
	    H1:8,H2:8,   ".", Mi1:8,Mi2:8, ".", S1:8,S2:8,
	    B0/binary>>, S) ->
    Date = {{list_to_integer("19" ++ [Y1, Y2]), 
	     list_to_integer([M1, M2]), 
	     list_to_integer([D1, D2])},
	    {list_to_integer([H1, H2]),
	     list_to_integer([Mi1, Mi2]),
	     list_to_integer([S1, S2])}},
    {B1,_} = split_at(B0, $;, S),
    {B1, Date};
parse_date1(<<Y1:8,Y2:8,Y3:8,Y4:8, ".", M1:8,M2:8,   ".", D1:8,D2:8, ".",
	    H1:8,H2:8,   ".", Mi1:8,Mi2:8, ".", S1:8,S2:8, 
	    B0/binary>>, S) ->
    Date = {{list_to_integer([Y1, Y2, Y3, Y4]), 
	     list_to_integer([M1, M2]), 
	     list_to_integer([D1, D2])},
	    {list_to_integer([H1, H2]),
	     list_to_integer([Mi1, Mi2]),
	     list_to_integer([S1, S2])}},
    {B1,_} = split_at(B0, $;, S),
    {B1, Date}.


string_plus_semi(B0, S) ->
    {B1, Str} = parse_string(B0, S),
    {B2, _} = split_at(B1, $;, S),
    {B2, Str}.


parse_string(B, S) ->
    <<"@", T/binary>> = B,
    parse_string(T, S, []).

parse_string(B, S, Acc) when size(B) < 2 ->
    case chunk(S) of
	{ok, Bin} ->
	    parse_string(list_to_binary([B,Bin]), S, Acc);
	eof ->
	    parse_string1(B, S, Acc)
    end;
parse_string(B, S, Acc) ->
    parse_string1(B, S, Acc).

%% parse_string1/3 is supposed to tail-recurse to parse_string/3.
%%
parse_string1(<<"@@", T/binary>>, S, Acc) ->
    parse_string(T, S, "@" ++ Acc);
parse_string1(<<"@", T/binary>>, _S, Acc) ->
    {T, lists:reverse(Acc)};
parse_string1(<<H:8, T/binary>>, S, Acc) ->
    parse_string(T, S, [H|Acc]).


%% bin2string(B) when size(B) > 15 ->
%%     binary_to_list(B, 1, 15);
%% bin2string(B) ->
%%     binary_to_list(B).


%%DESCRIPTION (99/06/05)
%%     An RCS file's contents are described by the grammar below.
%%
%%     The text is free format:  space,  backspace,  tab,  newline,
%%     vertical  tab, form feed, and carriage return (collectively,
%%     white space) have no significance except in  strings.   How-
%%     ever,  white  space cannot appear within an id, num, or sym,
%%     and an RCS file must end with a newline.
%%
%%     Strings are enclosed by @.  If a string  contains  a  @,  it
%%     must  be  doubled;  otherwise, strings can contain arbitrary
%%     binary data.
%%
%%     The meta syntax uses the following  conventions:  `|'  (bar)
%%     separates   alternatives;   `{'  and  `}'  enclose  optional
%%     phrases; `{' and `}*' enclose phrases that can  be  repeated
%%     zero  or  more times; `{' and '}+' enclose phrases that must
%%     appear at least once and can be repeated;  Terminal  symbols
%%     are in boldface; nonterminal symbols are in italics.
%%
%%     rcstext   ::=  admin {delta}* desc {deltatext}*
%%
%%     admin     ::=  head       {num};
%%                    { branch   {num}; }
%%                    access     {id}*;
%%                    symbols    {sym : num}*;
%%                    locks      {id : num}*;  {strict  ;}
%%                    { comment  {string}; }
%%                    { expand   {string}; }
%%                    { newphrase }*
%%
%%     delta     ::=  num
%%                    date       num;
%%                    author     id;
%%                    state      {id};
%%                    branches   {num}*;
%%                    next       {num};
%%                    { newphrase }*
%%
%%     desc      ::=  desc       string
%%
%%     deltatext ::=  num
%%                    log        string
%%                    { newphrase }*
%%                    text       string
%%
%%     num       ::=  {digit | .}+
%%
%%     digit     ::=  0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9
%%
%%     id        ::=  {num} idchar {idchar | num}*
%%
%%     sym       ::=  {digit}* idchar {idchar | digit}*
%%
%%     idchar    ::=  any visible graphic character except special
%%
%%     special   ::=  $ | , | . | : | ; | @
%%
%%     string    ::=  @{any character, with @ doubled}*@
%%
%%     newphrase ::=  id word* ;
%%
%%     word      ::=  id | num | string | :
%%
%%     Identifiers are case sensitive.  Keywords are in lower  case
%%     only.  The sets of keywords and identifiers can overlap.  In
%%     most environments RCS uses the ISO 8859/1 encoding:  visible
%%     graphic  characters are codes 041-176 and 240-377, and white
%%     space characters are codes 010-015 and 040.
%%
%%     Dates, which appear after the date keyword, are of the  form
%%     Y.mm.dd.hh.mm.ss, where Y is the year, mm the month (01-12),
%%     dd the day (01-31), hh  the  hour  (00-23),  mm  the  minute
%%     (00-59),  and  ss  the  second (00-60).  Y contains just the
%%     last two digits of the year  for  years  from  1900  through
%%     1999, and all the digits of years thereafter.  Dates use the
%%     Gregorian calendar; times use UTC.
%%
%%     The newphrase productions in the grammar  are  reserved  for
%%     future  extensions to the format of RCS files.  No newphrase
%%     will begin with any keyword already in use.
%%
%%     The delta nodes form a tree.  All nodes whose  numbers  con-
%%     sist  of  a  single pair (e.g., 2.3, 2.1, 1.3, etc.)  are on
%%     the trunk, and are linked through the next field in order of
%%     decreasing numbers.  The head field in the admin node points
%%     to the head of that sequence  (i.e.,  contains  the  highest
%%     pair).   The  branch  node  in  the admin node indicates the
%%     default branch (or revision) for most  RCS  operations.   If
%%     empty,  the  default  branch  is  the  highest branch on the
%%     trunk.
%%
%%     All delta nodes whose numbers consist  of  2n  fields  (n>2)
%%     (e.g.,  3.1.1.1, 2.1.2.2, etc.)  are linked as follows.  All
%%     nodes whose first  2n-1  number  fields  are  identical  are
%%     linked  through  the  next  field  in  order  of  increasing
%%     numbers.  For each  such  sequence,  the  delta  node  whose
%%     number  is  identical to the first 2n-2 number fields of the
%%     deltas on that sequence  is  called  the  branchpoint.   The
%%     branches  field  of a node contains a list of the numbers of
%%     the first nodes of all sequences for which  it  is  a  bran-
%%     chpoint.  This list is ordered in increasing numbers.
%%
%%     The following diagram shows an  example  of  an  RCS  file's
%%     organization.
%%
%%                                Head
%%                                  |
%%                                  |
%%                                  v                        / \
%%                              ---------                   /   \
%%        / \          / \      |       |      / \         /     \
%%       /   \        /   \     |  2.1  |     /   \       /       \
%%      /     \      /     \    |       |    /     \     /         \
%%     /1.2.1.3\    /1.3.1.1\   |       |   /1.2.2.2\   /1.2.2.1.1.1\
%%     ---------    ---------   ---------   ---------   -------------
%%         ^            ^           |           ^             ^
%%         |            |           |           |             |
%%         |            |           v           |             |
%%        / \           |       ---------      / \            |
%%       /   \          |       \  1.3  /     /   \           |
%%      /     \         ---------\     /     /     \-----------
%%     /1.2.1.1\                  \   /     /1.2.2.1\
%%     ---------                   \ /      ---------
%%         ^                        |           ^
%%         |                        |           |
%%         |                        v           |
%%         |                    ---------       |
%%         |                    \  1.2  /       |
%%         ----------------------\     /---------
%%                                \   /
%%                                 \ /
%%                                  |
%%                                  |
%%                                  v
%%                              ---------
%%                              \  1.1  /
%%                               \     /
%%                                \   /
%%                                 \ /

